# Use HERE Platform Service

> **Objectives**: Use the HERE Traffic API client in a Spark application to retrieve and visualize a real-time traffic flow on a map
>
> **Complexity**: Beginner
>
> **Time to complete**: 30 min
>
> **Prerequisites**: [Get your credentials](../topics/get-credentials.md), [Verify your credentials](../verify-credentials/README.md)
>
> **Source code**: <a href="real-time-traffic-tutorial.zip" download target="_blank">Download</a>


This tutorial demonstrates how to use the [HERE Traffic](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/overview) API client in a Spark application.
The [HERE Traffic](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/overview) API client is a part of the [Data Client Base Library]({{book.filtered.developerPortalUrl}}/documentation/data-client-base/dev_guide/index.html).
The Data Client Base Library provides Scala/Java libraries that you can use in your projects to access the HERE platform.
It offers a low-level, stateless and thread-safe programmatic access to [HERE platform Data APIs]({{book.filtered.developerPortalUrl}}/documentation/data-api/data_dev_guide/index.html)
as well as to the [HERE platform services](https://platform.here.com/services), such as `Geocode API`, `Routing API`, `RevGeocode API`, `HERE Traffic API` and so on.
For more information, see the [Data Client Base Library documentation]({{book.filtered.developerPortalUrl}}/documentation/data-client-base/dev_guide/index.html).

The HERE Traffic API is a RESTful API that:

- Provides access to real-time traffic flow data in JSON format, including information on the speed and jam factor for the region(s) defined in each request.
  The Traffic API v7 can also deliver additional data, such as the geometry of the road segments in relation to the flow.
- Provides aggregated information about traffic incidents in JSON format, including the type and location of each traffic incident, status, start and end time, and other relevant data. 
  This data is useful to dynamically optimize route calculations.

In this tutorial, we will get a snapshot of the real-time traffic flow, upload data to the versioned layer, and visualize the results on the map. 
Roads with a light load will be marked with a green line, with a medium load - with a yellow line, and with a high load - with a red line.

![Real Time Traffic](img/real-time-traffic.png)


The tutorial covers the following topics:

- [Set up a Maven project](#set-up-a-maven-project)
- [Use the Traffic API client](#use-the-traffic-api-client)
- [Implement the application](#implement-the-application)
- [Run the application locally](#run-the-application-locally)
- [Run the application on the platform](#run-the-application-on-the-platform)


## Set up a Maven project

In order to get real-time traffic conditions, you may download the source code at the beginning of the tutorial and store it in a folder of your choice,
or create a folder structure for your project from scratch as follows:

```
here-realtime-traffic-client
└── src
    └── main
        ├── java
        └── resources
        └── scala
```

You can do this with a single `bash` command:

```bash
mkdir -p here-realtime-traffic/src/main/{java,resources,scala}
```
The Maven POM file is similar to the one in the
[Verify Maven Settings](../verify-m2/README.md) tutorial, with the parent POM
and dependencies sections updated.

The Parent POM is `sdk-batch-bom_2.12` as this tutorial is designed to run Spark applications in both local and platform environments.

```
[snippet](pom.xml#parent)
```
The following dependencies are used:

- `com.here.platform.data.client.base:ols-traffic-v7_${scala.compat.version}` to use the HERE Traffic API client.
- `com.here.platform.data.client:local-support_${scala.compat.version}` to write data to a local catalog.
- `com.here.platform.data.client:spark-support_${scala.compat.version}` to write data to a catalog on the platform.
- `org.apache.spark:spark-core_${scala.compat.version}` to run a Java/Scala Spark Application.
- `com.here.platform.pipeline:pipeline-interface_${scala.compat.version}` to get information about input catalogs from the `PipelineContext`.
- `com.here.platform.location:location-integration-here-commons_${scala.compat.version}` containing the tools to calculate partition Tile IDs for certain geocoordinates. For more details, see [Calculate partition Tile IDs](../here-tile-resolver/README.md).
- `com.here.hrn:hrn_${scala.compat.version}` to parse catalog HRNs.

Dependencies:
```
[snippet](pom.xml#dependencies)
```

Once you have created the Maven project, the next step would be to write the code of the application and run it.

## Use the Traffic API client

As mentioned before, the tutorial shows how to write a simple Spark application that retrieves data from the HERE Traffic API and
visualize the data on the map. All the retrieved data is added to the versioned layer to visualize it with the [local data inspector]({{book.filtered.developerPortalUrl}}//documentation/open-location-platform-cli/user_guide/topics/local-data/local-layer-commands.html#catalog-layer-inspect). 
For more details, see [Local development and testing with CLI](../local-development-workflow/README.md#compile-and-run-locally).

HERE Real-Time Traffic comprises highly accurate data from multiple sources, including connected car probes, roadway sensors, and live operations centers.

Real-Time Traffic contains two types of traffic data:

- Flow - information about the speed of travel and congestion along a segment of a roadway. Flow data is updated every minute.
- Incidents - information about events that affect the flow of traffic or that may be important for drivers to know. Incident data is updated every two minutes.

Real-Time Traffic covers more than 70 countries globally. For full coverage information, see [Traffic coverage](https://developer.here.com/documentation/traffic-api/dev_guide/topics/coverage-information/traffic.html).

The HERE Traffic API provides the following endpoints:

- [`Real-Time Flow Information`](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/api-ref#operation/getFlow) - Get real-time traffic flow data for a geospatial area.
- [`Real-Time Incident Information`](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/api-ref#operation/getIncidents) - Get traffic incidents for a geospatial area.
- [`Real-Time Incident Information for the given ID`](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/api-ref#operation/getIncidentsById) - Get traffic incident for a given ID.

For more information about HERE Traffic API endpoints, see the [Traffic API Reference](https://developer.here.com/documentation/traffic-api/api-reference.html).

We will use the [`Real-Time Flow Information`](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/api-ref#operation/getFlow) endpoint to get real-time traffic flow data for a geospatial area.

A basic flow request consists of a geospatial filter (`in`) and the type of location referencing to return (`locationReferencing`). 
The response will contain the traffic flow information of the road segments that are within the requested area.

The geospatial filter `in` can be a `circle`, a `bounding box` or a `corridor`, each having its own format. 
In our application below, we will use a `circle` specified by a point with `latitude` and `longitude` and a `radius`: `in=circle:52.51652,13.37885;r=3000`. 
This parameter defines a circle with a center at the Brandenburg Gate and a radius of 3 km. For more details on the geospatial filtering types, see [Geospatial filtering](https://developer.here.com/documentation/traffic-api/dev_guide/topics/use-cases/filter-geospatial.html).

The `locationReferencing` can be one or more out of `tmc`, `olr`, `shape`.
A location reference describes a location. A location can be a particular point, curve, or two-dimensional shape on the surface of the earth. 
Quite often, when a location is used in an application, it refers to a particular anthropogenic or geographic feature, such as a road, building, mountain, or body of water.
In this tutorial, we will use the [`shape`](https://developer.here.com/documentation/traffic-api/dev_guide/topics/concepts/location-referencing.html#shape-points) location reference. This allows us to visualize the data on the map in a simple way. For more information on the location referencing types, see [Location referencing](https://developer.here.com/documentation/traffic-api/dev_guide/topics/concepts/location-referencing.html).

A request to this endpoint returns an array of flow items, each having a `location` and a `currentFlow` element.
Each location has a `length` in meters, a `description` that usually is a street name, and the available location references of the location, depending on which were requested by the `locationReferencing` parameter. In our case it's `latitudes` and `longitudes`.
The `currentFlow` element of each flow item carries the `speed`, `jam factor`, and `traversability` information of the flow item.

The code snippet below demonstrates how to get real-time flow information with the parameters described above:

{% codetabs name="Java", type="java" -%}

    FlowResponse response = trafficApi
            .getFlow()
            .withIn("circle:52.51652,13.37885;r=3000")
            .withLocationReferencing(Collections.singletonList("shape"))
            .withAdvancedFeatures(Collections.singletonList("deepCoverage"))
            .build()
            .toEntity();

{% language name="Scala", type="scala" -%}

    val response = trafficApi
      .getFlow(in = "circle:52.51652,13.37885;r=3000",
        locationReferencing = List("shape"),
        advancedFeatures = List("deepCoverage")
      )
      .toEntity()

{%- endcodetabs %}


After getting acquainted with the API client, we can proceed to the application review.

## Implement the application

Let's look at the implementation of this Spark application. In the code snippet below, you can see that
the main method logic is implemented in 3 methods:
- `getRealTimeTrafficFlow()`
- `prepareGeoJsonData()`
- `uploadPartitions()`

Let's take a closer look at the implementations of these methods.

The `getRealTimeTrafficFlow()` method is the main part of the application for obtaining real-time traffic flow. All other subsequent methods mentioned above process the data received from this method.
If you only want to get a real-time traffic flow without further processing, you just need to add the `com.here.platform.data.client.base:ols-traffic-v7_${scala.compat.version}` dependency mentioned above and add this method to your application.

The `getRealTimeTrafficFlow()` method creates a `HERE Traffic API client` and retrieves the real-time traffic flow using the parameters described at the beginning of this section.
The `HERE Traffic API` requires authentication. By default, the credentials stored in `$HOME/.here/credentials.properties` are used.
For more details, see the [Data Client Base Library documentation]({{book.filtered.developerPortalUrl}}/documentation/data-client-base/dev_guide/client/credentials.html).

The `prepareGeoJsonData()` method splits traffic data into partitions using [`HereTileResolver`]({{book.filtered.developerPortalUrl}}/documentation/location-library/api_reference_java/com/here/platform/location/integration/herecommons/geospatial/javadsl/HereTileResolver.html) (for details, see [Calculate partition Tile IDs](../here-tile-resolver/README.md)) and converts the data to GeoJSON per each tile.
Depending on the [`jamFactor`](https://developer.here.com/documentation/traffic-api/dev_guide/topics/concepts/flow.html#jam-factor) property, the data display style is different.
Roads with a light load will be marked with a green line, with a medium load - with a yellow line and with a high load - with a red line.
For more details, see [GeoJSON Data]({{book.filtered.developerPortalUrl}}/documentation/data-inspector-library/dev_guide/pages/howto-geojson-data.html) and [Style GeoJSON Visualization]({{book.filtered.developerPortalUrl}}/documentation/data-inspector-library/dev_guide/pages/style-geojson-visualization.html).
In this method, we use classes `LineString`, `Geometry`, `Feature`, `FeatureCollection` that represent GeoJSON. For more details, see the [Data Client Library documentation]({{book.filtered.developerPortalUrl}}/documentation/data-client-library/api_reference_scala/com/here/platform/data/client/model/geojson/index.html).

The `uploadPartitions()` method sends GeoJSON data to the versioned catalog using the [Data Client Library]({{book.filtered.developerPortalUrl}}/documentation/data-client-library/dev_guide/index.html).

For more information on what the application does, see the comments in the code below.

{% codetabs name="Java", type="java" -%}
[snippet](./src/main/java/RealTimeTrafficApiJava.java)
{% language name="Scala", type="scala" -%}
[snippet](./src/main/scala/RealTimeTrafficApiScala.scala)
{%- endcodetabs %}

Once the code is complete, you can prepare the resources and run the application.

## Run the application locally

To run the application, you need to prepare the resources - create a catalog with a versioned layer.

In this tutorial, we will run our application locally, therefore, it will be enough for us to create a local catalog. Since
they are contained in a local machine, they are not subject to naming conflicts within your realm and you can use any
name you want.

Although we do not use the input catalog in our tutorial,
we need to create it to fill the `input-catalog` field in the [config file](pipeline-config.conf), otherwise, you will get an error about an invalid catalog HRN.

Run the following OLP CLI command from the root of the tutorial folder to create a local input catalog:

```bash
olp local catalog create batch-catalog batch-catalog --summary "Input catalog" --description "Input catalog"
```

Run the following OLP CLI command from the root of the tutorial folder to create a local output catalog with the versioned layer:

```bash
olp local catalog create output-batch-catalog output-batch-catalog --config output-catalog-configuration.json
```

The structure of the [`output-catalog-configuration.json`](output-catalog-configuration.json) file is as follows:
{% codesnippet language="json" %}
[snippet](output-catalog-configuration.json)
{% endcodesnippet %}

Let's take a closer look at the parameters of the versioned layer. Property `"contentType": "application/geo+json"` means that the layer contains data in GeoJSON format.
Property: `scheme:heretile` means that the data is stored as a quad-tree. The data will be stored on `15` zoom-level. 
The local data inspector can visualize this tiling schema and GeoJSON without any additional rendering plugins.
After you have created the catalogs, run the application from the root of the downloaded tutorial using the following
command:

{% codetabs name="Java", type="java" -%}
mvn compile exec:java -q -D"exec.mainClass"="RealTimeTrafficApiJava" \
-Dhere.platform.data-client.endpoint-locator.discovery-service-env=local \
-Dspark.master=local[*] \
-Dpipeline-config.file=local-pipeline-config.conf
{% language name="Scala", type="scala" -%}
mvn compile exec:java -q -D"exec.mainClass"="RealTimeTrafficApiScala" \
-Dhere.platform.data-client.endpoint-locator.discovery-service-env=local \
-Dspark.master=local[*] \
-Dpipeline-config.file=local-pipeline-config.conf
{%- endcodetabs %}

The command has the following parameters:

- `exec.mainClass` – entry point to run your application.
- `here.platform.data-client.endpoint-locator.discovery-service-env=local` – configures the Data Client Library to only
  use local catalogs.
- `spark.master=local[*]` – configures a local Spark run with as many worker threads as there are logical cores on your machine.
- `pipeline-config.file=local-pipeline-config.conf` - configuration file with information about the input and output catalogs.

After the application finishes successfully, you can see the data that was added to the versioned layer in the console.

Once the GeoJSON data is published to the catalog, let's view the results using the following command:

```bash
olp local catalog layer inspect hrn:local:data:::output-batch-catalog realtime-traffic
```
This command will open the browser tab showing a real-time traffic flow:

![](img/real-time-traffic.png)

Now we can proceed to run this application on the platform.

## Run the application on the platform

Before running the application on the platform, we need to prepare all the resources for it.

We will run the application in the project. To do this, let's create a project using the OLP CLI.

```bash
olp project create $PROJECT_ID $PROJECT_NAME
```

The OLP CLI should return the following message:

```bash
Project YOUR_PROJECT_HRN has been created.
```

Save the `YOUR_PROJECT_HRN` value to the `$PROJECT_HRN` console variable to simplify command running.

Since we will be using the [HERE Traffic](https://platform.here.com/services/details/hrn:here:service::olp-here:traffic-api-7/overview) service, 
we need to link it to project, so that our HERE Traffic client could resolve the base url of the service. 
To do this, let's link the service to the project using the following command:

```bash 
olp project resource link $PROJECT_HRN hrn:here:service::olp-here:traffic-api-7
```

The next step is to create input and output catalogs in the project we have created before. The `--scope` parameter is
used to specify the project in the OLP CLI.

Note that the input catalog is not used in our application.
We need this catalog to specify the `input-catalog` property in the `pipeline-config.conf` file.

Let's execute the following command from the root of the tutorial folder to create an input catalog:

```bash
 olp catalog create $CATALOG_ID $CATALOG_NAME  --summary "Input catalog" --description "Input catalog" --scope $PROJECT_HRN
 ```

The OLP CLI should return the following message:

```bash
Catalog YOUR_INPUT_CATALOG_HRN has been created.
```

The output catalog should be created using the [`output-catalog-configuration.json`](output-catalog-configuration.json) file.
Run the following command from the root of the tutorial folder to create the output catalog:

```bash
 olp catalog create $CATALOG_ID $CATALOG_NAME --config output-catalog-configuration.json --scope $PROJECT_HRN
 ```

The next step is to configure the source code to use the output catalog we have created before.
Replace the `INPUT_CATALOG_HRN` and `OUTPUT_CATALOG_HRN` placeholders in the [`pipeline-config.conf`](pipeline-config.conf) file with the HRNs of the catalogs you created as described above.

The application has to be packaged with all dependencies in a Fat JAR to be deployed on the platform. The Java/Scala SDK
provides a `platform` profile to generate the Fat JAR. The following command uses the profile to generate the Fat JAR:

 ```bash
 mvn -Pplatform clean package
 ``` 

The `real-time-traffic-tutorial-<tutorial-version>-platform.jar` Fat JAR should be created in the **target** folder.

Now you have all the resources used by the application, and the application configured to use them and prepared to be
deployed on the platform.

Let's create a pipeline in the project scope using the following OLP CLI command:

```bash
 olp pipeline create $PIPELINE_ID --scope $PROJECT_HRN
```

The OLP CLI should return the following message:

```bash
Pipeline YOUR_PIPELINE_ID has been created.
```

The next step is to create a pipeline template. 

To create your own pipeline template, you need to specify the template
name, `batch` [runtime environment]({{book.filtered.devGuidesUrlBase}}/open-location-platform-cli/user_guide/topics/pipeline/environment-commands.html), since in this tutorial we run a Spark application and use the `batch` environment on the platform,
the Fat JAR created by the `mvn package -Pplatform` command, the main class, the project that the pipeline belongs to, and input
catalog IDs that are expected in the pipeline version configuration.


Let's create a pipeline template by running the following OLP CLI command from the root of the tutorial folder:

{% codetabs name="Java Application", type="bash" -%}
{% raw %}
olp pipeline template create traffic-template \
batch-3.0 target/real-time-traffic-tutorial-<tutorial-version>-platform.jar \
RealTimeTrafficApiJava --input-catalog-ids pipeline-config.conf \
--scope $PROJECT_HRN
{% endraw %}
{% language name="Scala Application", type="bash" -%}
{% raw %}
olp pipeline template create traffic-template \
batch-3.0 target/real-time-traffic-tutorial-<tutorial-version>-platform.jar \
RealTimeTrafficApiScala --input-catalog-ids pipeline-config.conf \
--scope $PROJECT_HRN
{% endraw %}
{%- endcodetabs %}


The next step is to create a pipeline version. 

Let's create a pipeline version in the project scope.

```bash
  olp pipeline version create test-spark-version $PIPELINE_ID $PIPELINE_TEMPLATE_ID pipeline-config.conf --scope $PROJECT_HRN
```

> #### Note
> If a billing tag is required in your realm, use the `--billing-tag <your-billing-tag>` parameter.


Once you have created the batch pipeline, pipeline template and pipeline version, you can run the application on the platform by activating the pipeline version.

To activate the pipeline version, run the following OLP CLI command:

```bash
  olp pipeline version activate $PIPELINE_ID $PIPELINE_VERSION_ID --scope $PROJECT_HRN
```

After the pipeline is activated, you need to wait until the pipeline reaches the `Completed` state.
For this purpose, execute the following OLP CLI command:

```bash
olp pipeline version wait $PIPELINE_ID $PIPELINE_VERSION_ID \
                --job-state completed --timeout 600 \
                --scope $PROJECT_HRN
```

After your pipeline has reached the `Completed` state, you can check the result of the application.

For this purpose, execute the following OLP CLI command:

```bash
olp catalog layer inspect $OUTPUT_CATALOG_HRN realtime-traffic
```
This command will open the browser tab showing a real-time traffic flow:

![](img/real-time-traffic.png)

## Conclusion

In this tutorial, you have learned how to use the HERE Traffic API client in the Java/Scala Spark application 
to get a snapshot of the real-time traffic flow and visualize it on a map.

## Further information

For more details on the topics covered in this tutorial, see the following sources:

- For more information about HERE Traffic API, see the 
  [HERE Traffic API v7 Developer Guide](https://developer.here.com/documentation/traffic-api/dev_guide/topics/concepts/flow.html#jam-factor)
- For more information about HERE Traffic API endpoints, see the
  [HERE Traffic API Reference](https://developer.here.com/documentation/traffic-api/api-reference.html)
- For more information on how to use Data Client Base in your project,
  see the [Data Client Base Library documentation]({{book.filtered.developerPortalUrl}}/documentation/data-client-base/dev_guide/client/credentials.html).
- For more information on how to develop Spark applications, see the
  [Apache Spark Programming Guides](https://spark.apache.org/docs/latest/quick-start.html).
- For more information about OLP CLI commands, see the
  [Command Line Interface Documentation]({{book.filtered.developerPortalUrl}}/documentation/open-location-platform-cli/user_guide/index.html).
- For more information about the Local Data Inspector, see the [documentation]({{book.filtered.developerPortalUrl}}/documentation/data-inspector-library/dev_guide/index.html).
- For more details about utilities connecting Spark with HERE platform data, see the [Data Client Library Documentation]({{book.filtered.developerPortalUrl}}/documentation/data-client-library/dev_guide/index.html).

