/*
 * Copyright (c) 2018-2023 HERE Europe B.V.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import akka.actor.ActorSystem;
import akka.stream.javadsl.Source;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.here.hrn.HRN;
import com.here.platform.data.client.base.javadsl.BaseClient;
import com.here.platform.data.client.base.javadsl.BaseClientJava;
import com.here.platform.data.client.base.ols.generated.javadsl.api.traffic.RealTimeTrafficApi;
import com.here.platform.data.client.base.ols.generated.scaladsl.model.traffic.CurrentFlow;
import com.here.platform.data.client.base.ols.generated.scaladsl.model.traffic.FlowItem;
import com.here.platform.data.client.base.ols.generated.scaladsl.model.traffic.FlowResponse;
import com.here.platform.data.client.base.ols.generated.scaladsl.model.traffic.JLocationShape;
import com.here.platform.data.client.engine.javadsl.DataEngine;
import com.here.platform.data.client.engine.javadsl.WriteEngine;
import com.here.platform.data.client.javadsl.NewPartition;
import com.here.platform.data.client.model.PendingPartition;
import com.here.platform.data.client.model.geojson.Feature;
import com.here.platform.data.client.model.geojson.FeatureCollection;
import com.here.platform.data.client.model.geojson.LineString;
import com.here.platform.data.client.spark.DataClientSparkContextUtils;
import com.here.platform.location.core.geospatial.GeoCoordinate;
import com.here.platform.location.integration.herecommons.geospatial.HereTileLevel;
import com.here.platform.location.integration.herecommons.geospatial.javadsl.HereTileResolver;
import com.here.platform.pipeline.PipelineContext;
import java.io.Serializable;
import java.util.*;
import java.util.concurrent.ExecutionException;

public class RealTimeTrafficApiJava {

  private static final String LAYER_ID = "realtime-traffic";

  public static void main(String[] args)
      throws JsonProcessingException, ExecutionException, InterruptedException {

    // get real time traffic flow
    FlowResponse flowResponse = getRealTimeTrafficFlow("circle:52.51652,13.37885;r=3000");

    // convert the retrieved data to GeoJson and split them in tiles
    List<PartitionData> partitionData = prepareGeoJsonData(flowResponse.getResults());

    // publish data to output catalog
    PipelineContext pipelineContext = new PipelineContext();
    uploadPartitions(pipelineContext.getConfig().getOutputCatalog(), partitionData);
  }

  /** Get real-time traffic flow from the RealTimeTrafficApi */
  private static FlowResponse getRealTimeTrafficFlow(String geo) {
    BaseClient client = BaseClientJava.instance();
    RealTimeTrafficApi trafficApi = new RealTimeTrafficApi(client);

    return trafficApi
        .getFlow()
        .withIn(geo)
        .withLocationReferencing(Collections.singletonList("shape"))
        .withAdvancedFeatures(Collections.singletonList("deepCoverage"))
        .build()
        .toEntity();
  }

  /** Convert the real-time traffic flow response to GeoJson and splitting them by tiles */
  private static List<PartitionData> prepareGeoJsonData(List<FlowItem> flowItems)
      throws JsonProcessingException {

    ObjectMapper objectMapper = new ObjectMapper();
    objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);

    List<PartitionData> customDataList = new ArrayList<>();
    Map<Long, List<Feature>> tileIdToFeatureList = new HashMap<>();

    // go through the flow items and create features from the data
    flowItems.forEach(
        flowItem -> {
          List<List<Double>> coordinates = new ArrayList<>();
          // get coordinates from the location to create the geometry from
          flowItem
              .getLocation()
              .getShape()
              .orElse(new JLocationShape.Builder().build())
              .getLinks()
              .stream()
              .flatMap(link -> link.getPoints().stream())
              .forEach(point -> coordinates.add(Arrays.asList(point.getLng(), point.getLat())));

          // create lineString geometry
          LineString lineString = new LineString.Builder().withCoordinates(coordinates).build();

          // create feature using the lineString geometry and flow properties, such as speed,
          // jamFactor, etc
          Feature feature =
              new Feature.Builder()
                  .withId(UUID.randomUUID().toString())
                  .withGeometry(lineString)
                  .withProperties(getGeoJsonProperties(flowItem.getCurrentFlow()))
                  .build();

          // find tile ID for the given coordinate and save feature per each tile ID
          lineString
              .getCoordinates()
              .forEach(
                  list -> {
                    double latitude = list.get(1);
                    double longitude = list.get(0);
                    GeoCoordinate coordinate = new GeoCoordinate(latitude, longitude);

                    long tileId =
                        new HereTileResolver(new HereTileLevel(15)).fromCoordinate(coordinate);
                    List<Feature> featureList =
                        tileIdToFeatureList.computeIfAbsent(tileId, (k) -> new ArrayList<>());
                    featureList.add(feature);
                  });
        });

    // create partition data from the features
    for (Map.Entry<Long, List<Feature>> entry : tileIdToFeatureList.entrySet()) {
      Long tileId = entry.getKey();
      FeatureCollection featureCollection =
          new FeatureCollection.Builder().withFeatures(entry.getValue()).build();
      PartitionData customData =
          new PartitionData(
              Long.toString(tileId), LAYER_ID, objectMapper.writeValueAsBytes(featureCollection));
      customDataList.add(customData);
    }
    return customDataList;
  }

  /** Upload data to the versioned layer using Data Client */
  public static void uploadPartitions(HRN catalog, List<PartitionData> partitions)
      throws ExecutionException, InterruptedException {

    ActorSystem actorSystem = DataClientSparkContextUtils.context().actorSystem();

    // create writeEngine for source catalog
    WriteEngine writeEngine = DataEngine.get(actorSystem).writeEngine(catalog);

    // parallelism defines how many parallel requests would be made to fetch the data
    int parallelism = 10;

    // create a list partitions to upload
    ArrayList<PendingPartition> partitionList = new ArrayList<>();
    partitions.forEach(
        partitionData -> {
          NewPartition newPartition =
              new NewPartition.Builder()
                  .withPartition(partitionData.partition)
                  .withData(partitionData.data)
                  .withLayer(LAYER_ID)
                  .build();
          partitionList.add(newPartition);
        });

    // upload partitions to the catalog
    writeEngine
        .publishBatch2(
            parallelism,
            Optional.of(Collections.singletonList(LAYER_ID)),
            Collections.emptyList(),
            Source.from(partitionList))
        .toCompletableFuture()
        .get();
  }

  /** Generate the real-time traffic flow properties for the Feature object */
  private static Map<String, Object> getGeoJsonProperties(CurrentFlow currentFlow) {

    Map<String, Object> properties = new HashMap<>();
    properties.put("style", generateFillColor(currentFlow.getJamFactor()));
    properties.put("speed", currentFlow.getSpeed().orElse(0.0));
    properties.put("speedUncapped", currentFlow.getSpeedUncapped().orElse(0.0));
    properties.put("freeFlow", currentFlow.getFreeFlow());
    properties.put("jamFactor", currentFlow.getJamFactor());
    properties.put("confidence", currentFlow.getConfidence().orElse(0.0));
    properties.put("traversability", currentFlow.getTraversability().orElse("empty"));

    return properties;
  }

  /**
   * Generate RGB encoded color. For high Jam Factor color have more red component, for low Jam
   * Factor - more green component. Each color component can have value between 0 and 255, blue
   * component not used.
   */
  private static Map<String, Object> generateFillColor(double jamFactor) {
    Map<String, Object> style = new HashMap<>();
    int green;
    int red;
    if (jamFactor <= 3) {
      green = 255;
      red = (int) (255 * (jamFactor * 10) / 100 * 2);
    } else {
      green = (int) (255 * ((100 - (jamFactor * 10)) / 100));
      red = 255;
    }
    style.put("color", "rgb(" + red + ", " + green + ", 0)");
    style.put("weight", "5");
    return style;
  }

  /** Class to store the metadata and real-time traffic flow data */
  @SuppressWarnings("serial")
  static class PartitionData implements Serializable {

    String partition;

    String layerId;

    byte[] data;

    PartitionData(String partition, String layerId, byte[] data) {
      this.partition = partition;
      this.layerId = layerId;
      this.data = data;
    }
  }
}
