/*
 * Copyright (c) 2018-2023 HERE Europe B.V.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import akka.stream.scaladsl.Source
import com.fasterxml.jackson.annotation.JsonInclude
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.scala.DefaultScalaModule
import com.here.hrn.HRN
import com.here.platform.data.client.base.ols.generated.codecs.traffic.JsonSupport._
import com.here.platform.data.client.base.ols.generated.scaladsl.api.traffic.RealTimeTrafficApi
import com.here.platform.data.client.base.ols.generated.scaladsl.model.traffic.{
  CurrentFlow,
  FlowItem,
  FlowResponse
}
import com.here.platform.data.client.base.scaladsl.BaseClient
import com.here.platform.data.client.engine.scaladsl.DataEngine
import com.here.platform.data.client.model.geojson.{Feature, FeatureCollection, LineString}
import com.here.platform.data.client.scaladsl.NewPartition
import com.here.platform.data.client.spark.DataClientSparkContextUtils.context.actorSystem
import com.here.platform.location.core.geospatial.GeoCoordinate
import com.here.platform.location.integration.herecommons.geospatial.{
  HereTileLevel,
  HereTileResolver
}
import com.here.platform.pipeline.PipelineContext

import java.util
import scala.collection.convert.ImplicitConversions.`map AsJavaMap`
import scala.collection.mutable.ListBuffer
import scala.concurrent.ExecutionContext.Implicits.global

object RealTimeTrafficApiScala {
  private val LAYER_ID = "realtime-traffic"

  def main(args: Array[String]): Unit = {

    // get real time traffic flow
    val flowResponse = getRealTimeTrafficFlow("circle:52.51652,13.37885;r=3000")

    // convert the retrieved data to GeoJson and split them in tiles
    val partitionData = prepareGeoJsonData(flowResponse.getResults)

    // publish data to output catalog
    val pipelineContext = new PipelineContext
    uploadPartitions(pipelineContext.config.outputCatalog, partitionData)

  }

  /**
    * Get real-time traffic flow from the RealTimeTrafficApi
    */
  def getRealTimeTrafficFlow(str: String): FlowResponse = {

    val client = BaseClient()
    val trafficApi = client.of[RealTimeTrafficApi]

    trafficApi
      .getFlow(
        in = str,
        locationReferencing = List("shape"),
        advancedFeatures = List("deepCoverage")
      )
      .toEntity()
  }

  /**
    * Convert the real-time traffic flow response to GeoJson
    * and splitting them by tiles
    */
  def prepareGeoJsonData(flowItems: util.List[FlowItem]): List[PartitionData] = {

    val objectMapper = new ObjectMapper()
    objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL)
    objectMapper.registerModule(new DefaultScalaModule)

    val customDataList = ListBuffer[PartitionData]()
    val tileIdToFeatureList = scala.collection.mutable.Map[Long, ListBuffer[Feature]]()

    // go through the flow items and create features from the data
    flowItems.forEach(item => {

      val coordinates = ListBuffer(ListBuffer[Double]())

      // get coordinates from the location to create the geometry from
      item.getLocation.shape.get.links
        .flatMap(link => link.points)
        .foreach(point => {
          coordinates += ListBuffer(point.getLng, point.getLat)
        })

      // create lineString geometry
      coordinates.remove(0)
      val lineString = LineString(
        coordinates = Option(coordinates.map(_.toList).toList)
      )

      // create feature using the lineString geometry and flow properties, such as speed,
      // jamFactor, etc
      val feature = Feature.apply(
        id = Option(java.util.UUID.randomUUID.toString),
        geometry = Option(lineString),
        properties = Option(getGeoJsonProperties(item.getCurrentFlow).toMap)
      )

      // find tile ID for the given coordinate and save feature per each tile ID
      lineString.getCoordinates.forEach(list => {

        val latitude = list.get(1)
        val longitude = list.get(0)
        val coordinate = GeoCoordinate.apply(latitude, longitude)

        val tileId = new HereTileResolver(HereTileLevel(15)).fromCoordinate(coordinate).value
        val featureList = tileIdToFeatureList.getOrElseUpdate(tileId, ListBuffer())

        featureList += feature
      })
    })

    // create partition data from the features
    tileIdToFeatureList
      .entrySet()
      .forEach(entry => {
        val tileId = entry.getKey
        val featureCollection = FeatureCollection.apply(features = entry.getValue.toList)

        val customData = PartitionData.apply(
          tileId.toString,
          LAYER_ID,
          objectMapper.writeValueAsBytes(featureCollection)
        )

        customDataList += customData

      })

    customDataList.toList
  }

  /**
    * Upload data to the versioned layer using Data Client
    */
  def uploadPartitions(catalog: HRN, partitionsData: Seq[PartitionData]): Unit = {
    // create writeEngine for a catalog
    val writeEngine = DataEngine().writeEngine(catalog)

    val partitionList = ListBuffer[NewPartition]()

    // create a list partitions to upload
    partitionsData.foreach(partitionsData => {
      val partition = NewPartition(
        partition = partitionsData.partition,
        layer = LAYER_ID,
        data = NewPartition.ByteArrayData(partitionsData.data)
      )

      partitionList += partition
    })

    // upload partitions to the catalog
    writeEngine.publishBatch2(parallelism = 10,
                              layerIds = Some(Seq(LAYER_ID)),
                              dependencies = Seq.empty,
                              partitions = Source.apply(partitionList.toList))
  }

  /**
    * Generate the real-time traffic flow properties for the Feature object
    */
  def getGeoJsonProperties(currentFlow: CurrentFlow): scala.collection.mutable.Map[String, Any] = {

    val properties = scala.collection.mutable.Map[String, Any]()

    properties += ("style" -> generateFillColor(currentFlow.getJamFactor))
    properties += ("speed" -> currentFlow.getSpeed.orElse(0.0))
    properties += ("speedUncapped" -> currentFlow.getSpeedUncapped.orElse(0.0))
    properties += ("freeFlow" -> currentFlow.getFreeFlow)
    properties += ("jamFactor" -> currentFlow.getJamFactor)
    properties += ("confidence" -> currentFlow.getConfidence.orElse(0.0))
    properties += ("traversability" -> currentFlow.getTraversability.orElse("empty"))

    properties

  }

  /**
    * Generate RGB encoded color. For high Jam Factor color have more red component, for low Jam
    * Factor - more green component. Each color component can have value between 0 and 255, blue
    * component not used.
    */
  private def generateFillColor(jamFactor: Double) = {
    val style = scala.collection.mutable.Map[String, Any]()
    var green = 0
    var red = 0
    if (jamFactor <= 3) {
      green = 255
      red = (255 * (jamFactor * 10) / 100 * 2).toInt
    } else {
      green = (255 * ((100 - (jamFactor * 10)) / 100)).toInt
      red = 255
    }
    val color = "rgb(" + red.toString + ", " + green.toString + ", 0)"
    style += ("color" -> color)
    style += ("weight" -> "5")
    style
  }

  /**
    * Class to store the metadata and real-time traffic flow data
    */
  case class PartitionData(partition: String, layer: String, data: Array[Byte])
}
